use avm1_tree as ast;
use nom::{IResult as NomResult, Needed};
use nom::{le_f32 as parse_le_f32, le_f64 as parse_le_f64, le_i16 as parse_le_i16, le_i32 as parse_le_i32, le_u16 as parse_le_u16, le_u8 as parse_u8};

use super::basic_data_types::{parse_bool_bits, parse_c_string, skip_bits};

#[derive(Debug, PartialEq, Eq)]
pub struct ActionHeader {
  pub action_code: u8,
  pub length: usize,
}

// TODO: Use nom::cond
pub fn parse_action_header(input: &[u8]) -> NomResult<&[u8], ActionHeader> {
  match parse_u8(input) {
    Ok((remaining_input, action_code)) => {
      if action_code < 0x80 {
        Ok((remaining_input, ActionHeader { action_code: action_code, length: 0 }))
      } else {
        parse_le_u16(remaining_input)
          .map(|(i, length)| (i, ActionHeader { action_code: action_code, length: length as usize }))
      }
    }
    Err(e) => Err(e),
  }
}

pub fn parse_goto_frame_action(input: &[u8]) -> NomResult<&[u8], ast::actions::GotoFrame> {
  do_parse!(
    input,
    frame: parse_le_u16 >>
    (ast::actions::GotoFrame {
      frame: frame as usize,
    })
  )
}

pub fn parse_get_url_action(input: &[u8]) -> NomResult<&[u8], ast::actions::GetUrl> {
  do_parse!(
    input,
    url: parse_c_string >>
    target: parse_c_string >>
    (ast::actions::GetUrl {
      url: url,
      target: target,
    })
  )
}

pub fn parse_store_register_action(input: &[u8]) -> NomResult<&[u8], ast::actions::StoreRegister> {
  do_parse!(
    input,
    register: parse_u8 >>
    (ast::actions::StoreRegister {
      register: register,
    })
  )
}

pub fn parse_constant_pool_action(input: &[u8]) -> NomResult<&[u8], ast::actions::ConstantPool> {
  do_parse!(
    input,
    constant_pool: length_count!(parse_le_u16, parse_c_string) >>
    (ast::actions::ConstantPool {
      constant_pool: constant_pool,
    })
  )
}

pub fn parse_wait_for_frame_action(input: &[u8]) -> NomResult<&[u8], ast::actions::WaitForFrame> {
  do_parse!(
    input,
    frame: parse_le_u16 >>
    skip_count: parse_u8 >>
    (ast::actions::WaitForFrame {
      frame: frame as usize,
      skip_count: skip_count as usize,
    })
  )
}

pub fn parse_set_target_action(input: &[u8]) -> NomResult<&[u8], ast::actions::SetTarget> {
  do_parse!(
    input,
    target_name: parse_c_string >>
    (ast::actions::SetTarget {
      target_name: target_name,
    })
  )
}

pub fn parse_goto_label_action(input: &[u8]) -> NomResult<&[u8], ast::actions::GoToLabel> {
  do_parse!(
    input,
    label: parse_c_string >>
    (ast::actions::GoToLabel {
      label: label,
    })
  )
}

pub fn parse_wait_for_frame2_action(input: &[u8]) -> NomResult<&[u8], ast::actions::WaitForFrame2> {
  do_parse!(
    input,
    skip_count: parse_u8 >>
    (ast::actions::WaitForFrame2 {
      skip_count: skip_count as usize,
    })
  )
}

#[derive(Debug, PartialEq, Eq)]
struct DefineFunction2Flags {
  pub preload_parent: bool,
  pub preload_root: bool,
  pub suppress_super: bool,
  pub preload_super: bool,
  pub suppress_arguments: bool,
  pub preload_arguments: bool,
  pub suppress_this: bool,
  pub preload_this: bool,
  pub preload_global: bool,
}

// TODO(demurgos): registerCount

pub fn parse_define_function2_action(input: &[u8]) -> NomResult<&[u8], ast::actions::DefineFunction2> {
  do_parse!(
    input,
    name: parse_c_string >>
    parameter_count: parse_le_u16 >>
    register_count: parse_u8 >>
    flags: bits!(do_parse!(
      preload_parent: call!(parse_bool_bits) >>
      preload_root: call!(parse_bool_bits) >>
      suppress_super: call!(parse_bool_bits) >>
      preload_super: call!(parse_bool_bits) >>
      suppress_arguments: call!(parse_bool_bits) >>
      preload_arguments: call!(parse_bool_bits) >>
      suppress_this: call!(parse_bool_bits) >>
      preload_this: call!(parse_bool_bits) >>
      apply!(skip_bits, 7) >>
      preload_global: call!(parse_bool_bits) >>
      (DefineFunction2Flags {
        preload_parent: preload_parent,
        preload_root: preload_root,
        suppress_super: suppress_super,
        preload_super: preload_super,
        suppress_arguments: suppress_arguments,
        preload_arguments: preload_arguments,
        suppress_this: suppress_this,
        preload_this: preload_this,
        preload_global: preload_global,
      })
    )) >>
    parameters: count!(map!(pair!(parse_u8, parse_c_string), |p: (u8, String)| ast::actions::define_function2::Parameter {register: p.0, name: p.1}), parameter_count as usize) >>
    code_size: parse_le_u16 >>
    body: take!(code_size as usize) >>
    (ast::actions::DefineFunction2 {
      name: name,
      preload_parent: flags.preload_parent,
      preload_root: flags.preload_root,
      suppress_super: flags.suppress_super,
      preload_super: flags.preload_super,
      suppress_arguments: flags.suppress_arguments,
      preload_arguments: flags.preload_arguments,
      suppress_this: flags.suppress_this,
      preload_this: flags.preload_this,
      preload_global: flags.preload_global,
      register_count: register_count as usize,
      parameters: parameters,
      body: body.to_vec(),
    })
  )
}

fn parse_catch_target(input: &[u8], catch_in_register: bool) -> NomResult<&[u8], ast::actions::r#try::CatchTarget> {
  if catch_in_register {
    parse_u8(input).map(|(i, v)| (i, ast::actions::r#try::CatchTarget::Register(v)))
  } else {
    parse_c_string(input).map(|(i, v): (_, String)| (i, ast::actions::r#try::CatchTarget::Variable(v)))
  }
}

pub fn parse_try_action(input: &[u8]) -> NomResult<&[u8], ast::actions::Try> {
  do_parse!(
    input,
    flags: bits!(do_parse!(
      apply!(skip_bits, 5) >>
      catch_in_register: parse_bool_bits >>
      finally_block: parse_bool_bits >>
      catch_block: parse_bool_bits >>
      ((catch_in_register, catch_block, finally_block))
    )) >>
    try_size: parse_le_u16 >>
    catch_size: parse_le_u16 >>
    finally_size: parse_le_u16 >>
    catch_target: call!(parse_catch_target, flags.0) >>
    try_body: take!(try_size as usize) >>
    catch_body: cond!(flags.1, map!(take!(catch_size as usize), |slice| slice.to_vec())) >>
    finally_body: cond!(flags.2, map!(take!(finally_size as usize), |slice| slice.to_vec())) >>
    (ast::actions::Try {
      r#try: try_body.to_vec(),
      catch_target: catch_target,
      catch: catch_body,
      finally: finally_body,
    })
  )
}

pub fn parse_with_action(input: &[u8]) -> NomResult<&[u8], ast::actions::With> {
  do_parse!(
    input,
    body_size: parse_le_i16 >>
    body: take!(body_size as usize) >>
    (ast::actions::With {
      with: body.to_vec(),
    })
  )
}

fn parse_action_value(input: &[u8]) -> NomResult<&[u8], ast::Value> {
  switch!(input, parse_u8,
   0 => map!(parse_c_string, |v: String| ast::Value::String(v)) |
   1 => map!(parse_le_f32, |v| ast::Value::Float32(v)) |
   2 => value!(ast::Value::Null) |
   3 => value!(ast::Value::Undefined) |
   4 => map!(parse_u8, |v| ast::Value::Register(v)) |
   5 => map!(parse_u8, |v| ast::Value::Boolean(v != 0)) |
   6 => map!(parse_le_f64, |v| ast::Value::Float64(v)) |
   7 => map!(parse_le_i32, |v| ast::Value::Sint32(v)) |
   8 => map!(parse_u8, |v| ast::Value::Constant(v as u16)) |
   9 => map!(parse_le_u16, |v| ast::Value::Constant(v))
  )
}

pub fn parse_push_action(input: &[u8]) -> NomResult<&[u8], ast::actions::Push> {
  let res = do_parse!(
    input,
    values: many1!(complete!(parse_action_value)) >>
    (ast::actions::Push {
      values: values,
    })
  );
  res
}

pub fn parse_jump_action(input: &[u8]) -> NomResult<&[u8], ast::actions::Jump> {
  do_parse!(
    input,
    branch_offset: parse_le_i16 >>
    (ast::actions::Jump {
      offset: branch_offset,
    })
  )
}

pub fn parse_get_url2_action(input: &[u8]) -> NomResult<&[u8], ast::actions::GetUrl2> {
  bits!(input, do_parse!(
    // TODO: Use switch! and value!
    send_vars_method: map!(
      take_bits!(u8, 2),
      |v| match v {
        0 => ast::actions::get_url2::SendVarsMethod::None,
        1 => ast::actions::get_url2::SendVarsMethod::Get,
        2 => ast::actions::get_url2::SendVarsMethod::Post,
        _ => panic!("Unexpected value for `send_vars_method`."),
      }
    ) >>
    apply!(skip_bits, 4) >>
    load_target: parse_bool_bits >>
    load_variables: parse_bool_bits >>
    (ast::actions::GetUrl2 {
      send_vars_method: send_vars_method,
      load_target: load_target,
      load_variables: load_variables,
    })
  ))
}

pub fn parse_define_function_action(input: &[u8]) -> NomResult<&[u8], ast::actions::DefineFunction> {
  do_parse!(
    input,
    name: parse_c_string >>
    parameter_count: parse_le_u16 >>
    parameters: count!(parse_c_string, parameter_count as usize) >>
    code_size: parse_le_u16 >>
    body: take!(code_size as usize) >>
    (ast::actions::DefineFunction {
      name: name,
      parameters: parameters,
      body: body.to_vec(),
    })
  )
}

pub fn parse_if_action(input: &[u8]) -> NomResult<&[u8], ast::actions::If> {
  do_parse!(
    input,
    branch_offset: parse_le_i16 >>
    (ast::actions::If {
      offset: branch_offset,
    })
  )
}

pub fn parse_goto_frame2_action(input: &[u8]) -> NomResult<&[u8], ast::actions::GotoFrame2> {
  do_parse!(
    input,
    flags: bits!(do_parse!(
      apply!(skip_bits, 6) >>
      scene_bias: parse_bool_bits >>
      play: parse_bool_bits >>
      ((scene_bias, play))
    )) >>
    scene_bias: cond!(flags.0, parse_le_u16) >>
    (ast::actions::GotoFrame2 {
      play: flags.1,
      scene_bias: match scene_bias {
        Some(b) => b as usize,
        None => 0,
      },
    })
  )
}

pub fn parse_action(input: &[u8]) -> NomResult<&[u8], ast::Action> {
  match parse_action_header(input) {
    Ok((remaining_input, ah)) => {
      if remaining_input.len() < ah.length {
        let action_header_length = input.len() - remaining_input.len();
        Err(::nom::Err::Incomplete(Needed::Size(action_header_length + ah.length)))
      } else {
        let result = match ah.action_code {
          0x04 => Ok((remaining_input, ast::Action::NextFrame)),
          0x05 => Ok((remaining_input, ast::Action::PrevFrame)),
          0x06 => Ok((remaining_input, ast::Action::Play)),
          0x07 => Ok((remaining_input, ast::Action::Stop)),
          0x08 => Ok((remaining_input, ast::Action::ToggleQuality)),
          0x09 => Ok((remaining_input, ast::Action::StopSounds)),
          0x0a => Ok((remaining_input, ast::Action::Add)),
          0x0b => Ok((remaining_input, ast::Action::Subtract)),
          0x0c => Ok((remaining_input, ast::Action::Multiply)),
          0x0d => Ok((remaining_input, ast::Action::Divide)),
          0x0e => Ok((remaining_input, ast::Action::Equals)),
          0x0f => Ok((remaining_input, ast::Action::Less)),
          0x10 => Ok((remaining_input, ast::Action::And)),
          0x11 => Ok((remaining_input, ast::Action::Or)),
          0x12 => Ok((remaining_input, ast::Action::Not)),
          0x13 => Ok((remaining_input, ast::Action::StringEquals)),
          0x14 => Ok((remaining_input, ast::Action::StringLength)),
          0x15 => Ok((remaining_input, ast::Action::StringExtract)),
          0x17 => Ok((remaining_input, ast::Action::Pop)),
          0x18 => Ok((remaining_input, ast::Action::ToInteger)),
          0x1c => Ok((remaining_input, ast::Action::GetVariable)),
          0x1d => Ok((remaining_input, ast::Action::SetVariable)),
          0x20 => Ok((remaining_input, ast::Action::SetTarget2)),
          0x21 => Ok((remaining_input, ast::Action::StringAdd)),
          0x22 => Ok((remaining_input, ast::Action::GetProperty)),
          0x23 => Ok((remaining_input, ast::Action::SetProperty)),
          0x24 => Ok((remaining_input, ast::Action::CloneSprite)),
          0x25 => Ok((remaining_input, ast::Action::RemoveSprite)),
          0x26 => Ok((remaining_input, ast::Action::Trace)),
          0x27 => Ok((remaining_input, ast::Action::StartDrag)),
          0x28 => Ok((remaining_input, ast::Action::EndDrag)),
          0x29 => Ok((remaining_input, ast::Action::StringLess)),
          0x2a => Ok((remaining_input, ast::Action::Throw)),
          0x2b => Ok((remaining_input, ast::Action::CastOp)),
          0x2c => Ok((remaining_input, ast::Action::ImplementsOp)),
          0x2d => Ok((remaining_input, ast::Action::FsCommand2)),
          0x30 => Ok((remaining_input, ast::Action::RandomNumber)),
          0x31 => Ok((remaining_input, ast::Action::MbStringLength)),
          0x32 => Ok((remaining_input, ast::Action::CharToAscii)),
          0x33 => Ok((remaining_input, ast::Action::AsciiToChar)),
          0x34 => Ok((remaining_input, ast::Action::GetTime)),
          0x35 => Ok((remaining_input, ast::Action::MbStringExtract)),
          0x36 => Ok((remaining_input, ast::Action::MbCharToAscii)),
          0x37 => Ok((remaining_input, ast::Action::MbAsciiToChar)),
          0x3a => Ok((remaining_input, ast::Action::Delete)),
          0x3b => Ok((remaining_input, ast::Action::Delete2)),
          0x3c => Ok((remaining_input, ast::Action::DefineLocal)),
          0x3d => Ok((remaining_input, ast::Action::CallFunction)),
          0x3e => Ok((remaining_input, ast::Action::Return)),
          0x3f => Ok((remaining_input, ast::Action::Modulo)),
          0x40 => Ok((remaining_input, ast::Action::NewObject)),
          0x41 => Ok((remaining_input, ast::Action::DefineLocal2)),
          0x42 => Ok((remaining_input, ast::Action::InitArray)),
          0x43 => Ok((remaining_input, ast::Action::InitObject)),
          0x44 => Ok((remaining_input, ast::Action::TypeOf)),
          0x45 => Ok((remaining_input, ast::Action::TargetPath)),
          0x46 => Ok((remaining_input, ast::Action::Enumerate)),
          0x47 => Ok((remaining_input, ast::Action::Add2)),
          0x48 => Ok((remaining_input, ast::Action::Less2)),
          0x49 => Ok((remaining_input, ast::Action::Equals2)),
          0x4a => Ok((remaining_input, ast::Action::ToNumber)),
          0x4b => Ok((remaining_input, ast::Action::ToString)),
          0x4c => Ok((remaining_input, ast::Action::PushDuplicate)),
          0x4d => Ok((remaining_input, ast::Action::StackSwap)),
          0x4e => Ok((remaining_input, ast::Action::GetMember)),
          0x4f => Ok((remaining_input, ast::Action::SetMember)),
          0x50 => Ok((remaining_input, ast::Action::Increment)),
          0x51 => Ok((remaining_input, ast::Action::Decrement)),
          0x52 => Ok((remaining_input, ast::Action::CallMethod)),
          0x53 => Ok((remaining_input, ast::Action::NewMethod)),
          0x54 => Ok((remaining_input, ast::Action::InstanceOf)),
          0x55 => Ok((remaining_input, ast::Action::Enumerate2)),
          0x60 => Ok((remaining_input, ast::Action::BitAnd)),
          0x61 => Ok((remaining_input, ast::Action::BitOr)),
          0x62 => Ok((remaining_input, ast::Action::BitXor)),
          0x63 => Ok((remaining_input, ast::Action::BitLShift)),
          0x64 => Ok((remaining_input, ast::Action::BitRShift)),
          0x65 => Ok((remaining_input, ast::Action::BitURShift)),
          0x66 => Ok((remaining_input, ast::Action::StrictEquals)),
          0x67 => Ok((remaining_input, ast::Action::Greater)),
          0x68 => Ok((remaining_input, ast::Action::StringGreater)),
          0x69 => Ok((remaining_input, ast::Action::Extends)),
          0x81 => map!(remaining_input, parse_goto_frame_action, |a| ast::Action::GotoFrame(a)),
          0x83 => map!(remaining_input, parse_get_url_action, |a| ast::Action::GetUrl(a)),
          0x87 => map!(remaining_input, parse_store_register_action, |a| ast::Action::StoreRegister(a)),
          0x88 => map!(remaining_input, parse_constant_pool_action, |a| ast::Action::ConstantPool(a)),
          0x8a => map!(remaining_input, parse_wait_for_frame_action, |a| ast::Action::WaitForFrame(a)),
          0x8b => map!(remaining_input, parse_set_target_action, |a| ast::Action::SetTarget(a)),
          0x8c => map!(remaining_input, parse_goto_label_action, |a| ast::Action::GotoLabel(a)),
          0x8d => map!(remaining_input, parse_wait_for_frame2_action, |a| ast::Action::WaitForFrame2(a)),
          0x8e => map!(remaining_input, parse_define_function2_action, |a| ast::Action::DefineFunction2(a)),
          0x8f => map!(remaining_input, parse_try_action, |a| ast::Action::Try(a)),
          0x94 => map!(remaining_input, parse_with_action, |a| ast::Action::With(a)),
          0x96 => map!(remaining_input, parse_push_action, |a| ast::Action::Push(a)),
          0x99 => map!(remaining_input, parse_jump_action, |a| ast::Action::Jump(a)),
          0x9a => map!(remaining_input, parse_get_url2_action, |a| ast::Action::GetUrl2(a)),
          0x9b => map!(remaining_input, parse_define_function_action, |a| ast::Action::DefineFunction(a)),
          0x9d => map!(remaining_input, parse_if_action, |a| ast::Action::If(a)),
          0x9e => Ok((remaining_input, ast::Action::Call)),
          0x9f => map!(remaining_input, parse_goto_frame2_action, |a| ast::Action::GotoFrame2(a)),
          _ => {
            Ok((
              &remaining_input[ah.length..],
              ast::Action::Unknown(ast::actions::UnknownAction { code: ah.action_code, data: (&remaining_input[..ah.length]).to_vec() })
            ))
          }
        };
        match result {
          Ok((remaining_input2, action)) => {
            // TODO: Check that we consumed at least ah.length
            Ok((remaining_input2, action))
          }
          a => a
        }
      }
    }
    Err(e) => Err(e),
  }
}

#[cfg(test)]
mod tests {
  use nom;

  use super::*;

  #[test]
  fn test_parse_push_action() {
    {
      let input = vec![0x04, 0x00, 0x07, 0x01, 0x00, 0x00, 0x00, 0x08, 0x02];
      let actual = parse_push_action(&input[..]);
      let expected = Ok((
        &[][..],
        ast::actions::Push {
          values: vec![
            ast::Value::Register(0),
            ast::Value::Sint32(1),
            ast::Value::Constant(2),
          ]
        },
      ));
      assert_eq!(actual, expected);
    }
    {
      let input = vec![0x00, 0x00];
      let actual = parse_push_action(&input[..]);
      let expected = Ok((
        &[][..],
        ast::actions::Push {
          values: vec![
            ast::Value::String(String::from("")),
          ]
        },
      ));
      assert_eq!(actual, expected);
    }
    {
      let input = vec![0x00, 0x01, 0x00];
      let actual = parse_push_action(&input[..]);
      let expected = Ok((
        &[][..],
        ast::actions::Push {
          values: vec![
            ast::Value::String(String::from("\x01")),
          ]
        },
      ));
      assert_eq!(actual, expected);
    }
  }

  #[test]
  fn test_parse_action_header() {
    {
      let input = vec![0b00000000, 0b00000000, 0b00000000, 0b00000000];
      assert_eq!(parse_action_header(&input[..]), Ok((&input[1..], ActionHeader { action_code: 0x00, length: 0 })));
    }
    {
      let input = vec![0b00000001, 0b00000000, 0b00000000, 0b00000000];
      assert_eq!(parse_action_header(&input[..]), Ok((&input[1..], ActionHeader { action_code: 0x01, length: 0 })));
    }
    {
      let input = vec![0b00010000, 0b00000000, 0b00000000, 0b00000000];
      assert_eq!(parse_action_header(&input[..]), Ok((&input[1..], ActionHeader { action_code: 0x10, length: 0 })));
    }
    {
      let input = vec![0b10000000, 0b00000000, 0b00000000, 0b00000000];
      assert_eq!(parse_action_header(&input[..]), Ok((&input[3..], ActionHeader { action_code: 0x80, length: 0 })));
    }
    {
      let input = vec![0b10000000, 0b00000001, 0b00000000, 0b00000000];
      assert_eq!(parse_action_header(&input[..]), Ok((&input[3..], ActionHeader { action_code: 0x80, length: 1 })));
    }
    {
      let input = vec![0b10000000, 0b00000000, 0b00000001, 0b00000000];
      assert_eq!(parse_action_header(&input[..]), Ok((&input[3..], ActionHeader { action_code: 0x80, length: 256 })));
    }
  }

  #[test]
  fn test_parse_action() {
    {
      let input = vec![0b00000001, 0b00000000, 0b00000000, 0b00000000];
      assert_eq!(
        parse_action(&input[..]),
        Ok((&input[1..], ast::Action::Unknown(ast::actions::UnknownAction { code: 0x01, data: Vec::new() })))
      );
    }
    {
      let input = vec![0b10000000, 0b00000001, 0b00000000, 0b00000011];
      assert_eq!(
        parse_action(&input[..]),
        Ok((&input[4..], ast::Action::Unknown(ast::actions::UnknownAction { code: 0x80, data: vec![0x03] })))
      );
    }
    {
      let input = vec![0b10000000, 0b00000010, 0b00000000, 0b00000011];
      assert_eq!(
        parse_action(&input[..]),
        Err(::nom::Err::Incomplete(nom::Needed::Size(5)))
      );
    }
  }
}
